# -*- coding: utf-8 -*-

module Sinatra::Browse
  class ParameterType
    attr_reader :name
    attr_reader :default
    attr_reader :validators
    attr_reader :description

    def self.inherited(subclass)
      subclass.class_eval do
        # Initialize the validator hash
        @validator_declarations ||= {}

        #
        # Global validators
        #
        validator(:in) { |possible_values| possible_values.member?(@value) }

        # We need a to_s here because the user should be allowed to define dependencies
        # using symbols while the actual keys of the params hash are strings
        validator(:depends_on) { |dep| @params.has_key?(dep.to_s) }
      end
    end

    def initialize(name, map)
      @name = name
      @default = map.delete(:default)

      @description = map.delete(:description) || map.delete(:desc)

      @transform = map.delete(:transform)
      @transform = @transform.to_proc if @transform

      @required = !! map[:required]
      @on_error = map.delete(:on_error)

      @validators = []

      map.each do |key, value|
        if val_blk = validator_declarations[key]
          @validators << Validator.new(
            name: key,
            criteria: map[key],
            validation_blk: val_blk
          )
        end
      end

    end

    def default
      @default.is_a?(Proc) ? @default.call : @default
    end

    def default_set?
      ! @default.nil?
    end

    def required?
      @required
    end

    def validate(params)
      @validators.each do |v|
        return false, build_error_hash(v.name, v.value) unless v.validate(self.name, params)
      end

      true
    end

    def transform(value)
      @transform ? @transform.call(value) : value
    end

    def coerce(value)
      raise NotImplementedError
    end

    def build_error_hash(reason, value)
      {
        reason: reason,
        parameter: self.name,
        value: value,
        on_error: @on_error
      }
    end

    def type
      type_string = self.class.to_s.split("::").last
      type_string.to_sym
    end

    def to_hash(options = {})
      h = {
        name: @name,
        type: type,
        required: required?,
      }

      if @default
        h[:default] = if @default.is_a?(Proc) && options[:noprocs]
          "dynamically generated"
        else
          @default
        end
      end

      @validators.each { |v| h[v.name.to_sym] = v.criteria }

      h
    end

    #
    # DSL
    #

    def self.validator(name, &blk)
      @validator_declarations[name] = blk
    end

    def self.validator_declarations
      @validator_declarations
    end

    def validator_declarations
      self.class.validator_declarations
    end

  end
end
